/*
 * Lexical analyzer
 */


#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>

#include "token.h"
#include "error.h"

#define MAX_STRLEN 255
#define MAX_LINE_LEN 238



// dieses Schweine-Macro kann Seiteneffekte haben...!
#define isIdentifierStart(c) (isalpha(c) || (c) == '_')
//#define isIdentifierPart(c) (isalpha(c) || isdigit(c) || (c) == '_')

// this must be a function instead of a macro
// otherwise isIdentifierPart(nextChar()) has side effects!
static int isIdentifierPart(char c)
{
  return isalnum(c) || c == '_';
}


// ---- multi-char operators ----

char *mc_ops[] = { "<>", "<=", ">=", "<<", ">>", ">>>" };


// ---- keywords ----

char *keywords[] = {
  "AND", "CASE", "CLEAR", "DATA", "DEF", "DIM", "DIV", "ELSE", "END",
  "ENDCASE", "ENDIF", "ENDPROC", "ENDWHILE", "EOR", "FALSE", "FN", "FOR",
  "IF", "INPUT", "LOAD", "LVAR", "MOD", "NEXT", "NOT", "OF", "OR",
  "OTHERWISE", "PRINT", "PROC", "QUIT", "REM", "REPEAT", "RUN", "SPC",
  "STEP", "TAB", "THEN", "TO", "TRUE", "UNTIL", "WHEN", "WHILE",
  NULL
};

const char *pos;   // points to the current char
int ttype;   // current token type



#define curr_char (*pos)

#define EOL '\0'  // end of line is NULL character (not newline)


char sval[256];
float nval;


void lex_start(const char *s)
{
  pos = s;
}


static char nextChar()
{
  // don't move past end of line
  if(*pos) pos++;
  return *pos;
}


/*
 * Reads a quoted string and stores it in s.
 */
static void getLiteral(char *s) {
  int i = 0;
  
  // curr_char == '"'

  while(1 /* true */) {
    nextChar();

    if(curr_char == EOL)
      error(9, "Missing \"");

    if(curr_char == '"')
      if(nextChar() != '"')
	break;

    // append curr_char
    if(i >= MAX_STRLEN)
      error(19, "String too long");
    s[i++] = curr_char;
  }
  s[i] = 0;
}


/**
 * Reads an identifier.
 */
static void getIdentifier(char *s) {
  int i = 0;
  
  s[i++] = curr_char;

  while(isIdentifierPart(nextChar())) {
    if(i >= MAX_STRLEN)
      error(19, "String too long");
    s[i++] = curr_char;
  }

  // append % or $ for integers and strings
  if(curr_char == '%' || curr_char == '$') {
    if(i >= MAX_STRLEN)
      error(19, "String too long");
    s[i++] = curr_char;
    nextChar();
  }
  s[i] = 0;
}


/*
 * Reads a number.
 * Currently no sign or exponent.
 */
static float getNumeric() {
  float val = curr_char - '0';
  
  while(isdigit(nextChar()))
    val = val * 10 + curr_char - '0';

  if(curr_char == '.') {
    float dec = 10;
    while(isdigit(nextChar())) {
      val += (curr_char - '0') / dec;
      dec *= 10;
    }
  }
  
  return val;
}


/*
 * Searches a keyword and returns its token or -1 if not found.
 */
static int lookup_keyword(const char *s)
{
  int i;

  for(i = 0; keywords[i]; i++)
    if(strcmp(s, keywords[i]) == 0)
      return i + 0x100;

  return -1;
}



int nextToken()
{
  // skip whitespace (space or tab)
  while(curr_char == ' ' || curr_char == '\t')
    nextChar();

  if(curr_char == '\"') {
    getLiteral(sval);
    ttype = TT_STRING;
  }
  else if(isIdentifierStart(curr_char)) {
    getIdentifier(sval);
    ttype = lookup_keyword(sval);
    // if it's no keyword then it's an identifier
    if(ttype == -1)
      ttype = TT_IDENTIFIER;
  }
  else if(isdigit(curr_char)) {
    nval = getNumeric();
    ttype = TT_NUMBER;
  }
  else if(curr_char == '\0')
    ttype = TT_EOL;
  else {
    ttype = curr_char;
    nextChar();
  }

  return ttype;
}


char *tokenToString(int token)
{
  static char buf[4];

  switch(token) {
  case TT_EOL:
    return "end of line";
  case TT_NUMBER:
    return "number";
  case TT_STRING:
    return "string";
  case TT_IDENTIFIER:
    return "identifier";
  default:
    if(token >= 0x100)
      return keywords[token - 0x100];
    else if(token >= 0x80)
      return mc_ops[token - 0x80];
    else if(token >= 0x80)
      return mc_ops[token - 0x80];
    else {
      sprintf(buf, "'%c'", token);
      return buf;
    }
  }
}

